#include <cassert>
#include <string>
#include <ctime>
#include <cmath>
#include <deque>
#include <limits>
#include "somt.h"
#include "mex.h"


#define WARP_SIZE 32
#define DEQUE_SIZE 10
#include "unbalanced.h"

using std::cout;
using std::endl;
using std::deque;



#define NOTDEBUGFLAG 



class mystream : public std::streambuf
{
protected:
virtual std::streamsize xsputn(const char *s, std::streamsize n) { mexPrintf("%.*s", n, s); return n; }
virtual int overflow(int c=EOF) { if (c != EOF) { mexPrintf("%.1s", &c); } return 1; }
};
class scoped_redirect_cout
{
public:
	scoped_redirect_cout() { old_buf = std::cout.rdbuf(); std::cout.rdbuf(&mout); }
	~scoped_redirect_cout() { std::cout.rdbuf(old_buf); }
private:
	mystream mout;
	std::streambuf *old_buf;
};
static scoped_redirect_cout mycout_redirect;



void mexFunction(int nlhs, mxArray *plhs[], int nrhs, mxArray const *prhs[]) {
  if ( nrhs!=9 ){
    mexErrMsgIdAndTxt("error:error",
                      "wrong number of inputs.");
  }
  
  int num_devices;
  cudaGetDeviceCount(&num_devices);
  if (num_devices > 1) {
    cout << "Multiple devices detected" << endl;
    cout << "you have" << endl;
    for (int i = 0; i < num_devices; i++) {
      cudaDeviceProp properties;
      cudaGetDeviceProperties(&properties, i);
      cout << "Device " << i << ": " << properties.name << endl;
    }
    cudaSetDevice(0);
    cout << "Choosing device: ";
    cudaDeviceProp properties;
    cudaGetDeviceProperties(&properties, 0);
    cout << properties.name << endl;
  }
  
  const int d = 2;
  float gamma;
  float mu;
  float nu;
  float tau;
  int max_iter;
  bool use_smart_interval;
  int norm_type;
  
  gamma = float(*mxGetPr(prhs[2]));
  mu = float(*mxGetPr(prhs[3]));
  nu = float(*mxGetPr(prhs[4]));
  tau = float(*mxGetPr(prhs[5]));
  max_iter = int(*mxGetPr(prhs[6]));
  use_smart_interval = mxIsLogicalScalarTrue( prhs[7] );
  norm_type = int(*mxGetPr(prhs[8]));


  
  if ( !(nlhs==1 || nlhs==5) ) {
    mexErrMsgIdAndTxt("error:error",
                      "output should be dist or [dist m Phi t_rho0 t_rho1]");
  }

  const int n = mxGetDimensions(prhs[0])[0];
  const float dx = 1.f / float(n - 1);
  
  const double* rho0_double = mxGetPr(prhs[0]);
  const double* rho1_double = mxGetPr(prhs[1]);
  
  float* rho0 = new float[n*n];
  float* rho1 = new float[n*n];
  
  //convert input to float
  for (int i = 0; i < n; i++) {
    for (int j = 0; j < n; j++) {
      rho0[i + n*j] = float(rho0_double[i + n*j]);
      rho1[i + n*j] = float(rho1_double[i + n*j]);
    }
  }

  float rho0_max, rho1_max;
  rho0_max = -1.0f*std::numeric_limits<float>::infinity(); //-inf
  rho1_max = -1.0f*std::numeric_limits<float>::infinity(); //-inf
  for (int ii=0; ii<n*n; ii++) {
    if (rho0_max < rho0[ii])
      rho0_max = rho0[ii];
    if (rho1_max < rho1[ii])
      rho1_max = rho1[ii]; 
  }

  
  if ( !( n==1 || n==2 || n==4 || n==8 || n==16 || n==32 || n==64 || n==128 || n==256) ) {
    mexErrMsgIdAndTxt("error:error",
                      "n has to be a power of 2");
  }
  
  
  
  int threads_per_block;
  if (n==1) {
    threads_per_block = 1;
  } else if (n==2) {
    threads_per_block = 4;
  } else if (n==4) {
    threads_per_block = 16;
  } else if (n==8) {
    threads_per_block = 64;
  } else if (n==16) {
    threads_per_block = 128;
  } else if (n==32) {
    threads_per_block = 128;
  } else if (n==64) {
    threads_per_block = 256;
  } else if (n==128) {
    threads_per_block = 256;
  } else if (n==256) {
    threads_per_block = 256;
  } else {
    mexErrMsgIdAndTxt("error:error",
                      "unknown error");
  }
  
  



  float* t_rho0 = new float[n*n];
  float* t_rho1 = new float[n*n];
  float* t_rho0_temp = new float[n*n];
  float* t_rho1_temp = new float[n*n];  

  for (int ii=0; ii<n*n; ii++) {
    t_rho0[ii] = rho0[ii];
    t_rho1[ii] = rho1[ii];
  }


  //initialize m and Phi to be 0
  float* m = new float[n*n*d];
  memset(m, 0, sizeof(float)*n*n*d);
  float* Phi = new float[n*n];
  memset(Phi, 0, sizeof(float)*n*n);
  float* m_prev = new float[n*n*d];
  float* Phi_prev = new float[n*n];
  float* t_rho0_prev = new float[n*n];
  float* t_rho1_prev = new float[n*n];


  

  //---------------------------------------------------------------------
  //start error checking
  //thread_per_block must divide number of points
  assert((n*n) % threads_per_block == 0);
  if (n*n / threads_per_block > 65535) {
    cout << "maximum block number exceeded" << endl;
    cout << "threads_per_block should be bigger" << endl;
    assert(false);
  }
  assert((n*n)%1024==0); //required for parallel reduction
  //end error checking
  //---------------------------------------------------------------------

  //create CUDA pointers
  float* d_rho0;
  float* d_rho1;
  float* d_m;
  float* d_m_temp;
  float* d_Phi;
  float* d_t_rho0;
  float* d_t_rho0_temp;
  float* d_t_rho1;
  float* d_t_rho1_temp;
  float* d_min0;
  float* d_max0;
  float* d_sum0;
  float* d_min1;
  float* d_max1;
  float* d_sum1;

  cout << "Allocating GPU memory" << endl;
  err_chk(cudaMalloc((void**)&d_rho0, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_rho1, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_m, sizeof(float)*n*n*d));
  err_chk(cudaMalloc((void**)&d_m_temp, sizeof(float)*n*n*d));
  err_chk(cudaMalloc((void**)&d_Phi, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_rho0, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_rho0_temp, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_rho1, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_rho1_temp, sizeof(float)*n*n));

  err_chk(cudaMalloc((void**)&d_min0, sizeof(float)));
  err_chk(cudaMalloc((void**)&d_max0, sizeof(float)));
  err_chk(cudaMalloc((void**)&d_sum0, sizeof(float)));
  err_chk(cudaMalloc((void**)&d_min1, sizeof(float)));
  err_chk(cudaMalloc((void**)&d_max1, sizeof(float)));
  err_chk(cudaMalloc((void**)&d_sum1, sizeof(float)));


  cout << "copying memory from host to GPU" << endl;
  err_chk(cudaMemcpy(d_m, m, sizeof(float)*n*n*d, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_Phi, Phi, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_rho0, rho0, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_rho1, rho1, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_t_rho0, t_rho0, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_t_rho1, t_rho1, sizeof(float)*n*n, cudaMemcpyHostToDevice));

  float* theta_min0;
  float* theta_max0;
  float* theta_mid0;
  float* sum_min0;
  float* sum_max0;
  float* sum0;
  float* sum0_2;
  float* theta_min1;
  float* theta_max1;
  float* theta_mid1;
  float* sum_min1;
  float* sum_max1;
  float* sum1;
  float* sum1_2;

  cudaMallocHost((void**)&theta_min0, sizeof(float));
  cudaMallocHost((void**)&theta_max0, sizeof(float));
  cudaMallocHost((void**)&theta_mid0, sizeof(float));
  cudaMallocHost((void**)&sum_min0, sizeof(float));
  cudaMallocHost((void**)&sum_max0, sizeof(float));  
  cudaMallocHost((void**)&sum0, sizeof(float));
  cudaMallocHost((void**)&sum0_2, sizeof(float));
  cudaMallocHost((void**)&theta_min1, sizeof(float));
  cudaMallocHost((void**)&theta_max1, sizeof(float));
  cudaMallocHost((void**)&theta_mid1, sizeof(float));
  cudaMallocHost((void**)&sum_min1, sizeof(float));
  cudaMallocHost((void**)&sum_max1, sizeof(float));  
  cudaMallocHost((void**)&sum1, sizeof(float));
  cudaMallocHost((void**)&sum1_2, sizeof(float));


  deque<float> theta_deque0;
  deque<float> theta_deque1;

  const int num_streams = 3;
  cudaStream_t streams[num_streams];
  cudaStreamCreate(&streams[0]);
  cudaStreamCreate(&streams[1]);
  cudaStreamCreate(&streams[2]);

  cout << "starting iteration" << endl;
  clock_t begin = clock();
  for (int k = 0; k < max_iter; k++) {
  
    //----------------------------------------
    //Start rho updates
    const float epsilon = 1e-6;
    
    rhoStep0 <<< n*n / threads_per_block, threads_per_block, 0,  streams[0] >>> (d_Phi, d_t_rho0, d_t_rho0_temp, nu, n);
    rhoStep1 <<< n*n / threads_per_block, threads_per_block, 0,  streams[1] >>> (d_Phi, d_t_rho1, d_t_rho1_temp, nu, n);
    
    //compute theta interval
    if (!use_smart_interval || k<DEQUE_SIZE) {
      *theta_min0 = 1.0f*std::numeric_limits<float>::infinity(); //inf
      *theta_max0 = -1.0f*std::numeric_limits<float>::infinity(); //-inf
      err_chk(cudaMemcpyAsync(d_min0, theta_min0, sizeof(float), cudaMemcpyHostToDevice, streams[0]));
      err_chk(cudaMemcpyAsync(d_max0, theta_max0, sizeof(float), cudaMemcpyHostToDevice, streams[0]));
      cuda_parallel_minmax <<<(n*n/1024), 1024, 0, streams[0]>>> (d_t_rho0, d_max0, d_min0);
      
      
      *theta_min1 = std::numeric_limits<float>::infinity(); //inf
      *theta_max1 = -1.0f*std::numeric_limits<float>::infinity(); //-inf
      err_chk(cudaMemcpyAsync(d_min1, theta_min1, sizeof(float), cudaMemcpyHostToDevice, streams[1]));
      err_chk(cudaMemcpyAsync(d_max1, theta_max1, sizeof(float), cudaMemcpyHostToDevice, streams[1]));
      cuda_parallel_minmax <<<(n*n/1024), 1024, 0, streams[1]>>> (d_t_rho1, d_max1, d_min1);
      
      
      err_chk(cudaMemcpyAsync(theta_max0, d_max0, sizeof(float), cudaMemcpyDeviceToHost, streams[0]));
      err_chk(cudaMemcpyAsync(theta_min0, d_min0, sizeof(float), cudaMemcpyDeviceToHost, streams[0]));
      err_chk(cudaMemcpyAsync(theta_max1, d_max1, sizeof(float), cudaMemcpyDeviceToHost, streams[1]));
      err_chk(cudaMemcpyAsync(theta_min1, d_min1, sizeof(float), cudaMemcpyDeviceToHost, streams[1]));
      
      if (norm_type == 1) {
        mUpdate_l1 << < n*n / threads_per_block, threads_per_block, 0, streams[2] >> > (d_Phi, d_m, d_m_temp, n, dx, mu);
      } else if (norm_type == 2) {
        mUpdate_l2 << < n*n / threads_per_block, threads_per_block, 0,  streams[2] >> > (d_Phi, d_m, d_m_temp, n, dx, mu);
      } else {
        assert(false);
      }
      
      cudaStreamSynchronize(streams[0]);
      *theta_min0 -= rho0_max;
      *sum_min0 = gamma;
      *sum_max0 = 0.0f;
      cudaStreamSynchronize(streams[1]);
      *theta_min1 -= rho1_max;
      *sum_min1 = gamma;
      *sum_max1 = 0.0f;
      
    } else {
      #ifndef NOTDEBUGFLAG 
      cout << "computing smart interval0" << endl;
      cout << "computing smart interval1" << endl;
      #endif
      get_interval(theta_deque0, *theta_min0, *theta_max0);
      get_interval(theta_deque1, *theta_min1, *theta_max1);
      //check if theta_min0 is valid
      //check if theta_min1 is valid
      *sum0 = 0.0f;
      *sum1 = 0.0f;
      err_chk(cudaMemcpyAsync(d_sum0, sum0, sizeof(float), cudaMemcpyHostToDevice, streams[0]));
      err_chk(cudaMemcpyAsync(d_sum1, sum1, sizeof(float), cudaMemcpyHostToDevice, streams[1]));
      parallel_sum <<<(n*n/1024),1024,0,streams[0]>>> (d_rho0, d_t_rho0, *theta_min0, d_sum0);
      parallel_sum <<<(n*n/1024),1024,0,streams[1]>>> (d_rho1, d_t_rho1, *theta_min1, d_sum1);
      err_chk(cudaMemcpyAsync(sum0, d_sum0, sizeof(float), cudaMemcpyDeviceToHost, streams[0]));
      err_chk(cudaMemcpyAsync(sum1, d_sum1, sizeof(float), cudaMemcpyDeviceToHost, streams[1]));
      
      *sum0_2 = 0.0f;
      *sum1_2 = 0.0f;
      
      cudaStreamSynchronize(streams[0]);
      if (*sum0 < gamma) { //theta_min0 was not valid
        *theta_max0 = *theta_min0;
        *sum_max0 = *sum0;
        *theta_min0 = std::numeric_limits<float>::infinity(); //inf
        err_chk(cudaMemcpyAsync(d_min0, theta_min0, sizeof(float), cudaMemcpyHostToDevice, streams[0]));
        cuda_parallel_min <<<(n*n/1024),1024,0,streams[0]>>> (d_t_rho0, d_min0);
        err_chk(cudaMemcpyAsync(theta_min0, d_min0, sizeof(float), cudaMemcpyDeviceToHost, streams[0]));
      } else {
        //theta_min0 = theta_min0;
        *sum_min0 = *sum0;

        err_chk(cudaMemcpyAsync(d_sum0, sum0_2, sizeof(float), cudaMemcpyHostToDevice, streams[0]));
        parallel_sum <<<(n*n/1024),1024,0,streams[0]>>> (d_rho0, d_t_rho0, *theta_max0, d_sum0);
        err_chk(cudaMemcpyAsync(sum0_2, d_sum0, sizeof(float), cudaMemcpyDeviceToHost, streams[0]));
      }
      
      cudaStreamSynchronize(streams[1]);
      if (*sum1 < gamma) { //theta_min1 was not valid
        *theta_max1 = *theta_min1;
        *sum_max1 = *sum1;
        *theta_min1 = std::numeric_limits<float>::infinity(); //inf
        err_chk(cudaMemcpyAsync(d_min1, theta_min1, sizeof(float), cudaMemcpyHostToDevice, streams[1]));
        cuda_parallel_min <<<(n*n/1024),1024,0,streams[1]>>> (d_t_rho1, d_min1);
        err_chk(cudaMemcpyAsync(theta_min1, d_min1, sizeof(float), cudaMemcpyDeviceToHost, streams[1]));
      } else {
        //*theta_min1 = *theta_min1;
        *sum_min1 = *sum1;

        err_chk(cudaMemcpyAsync(d_sum1, sum1_2, sizeof(float), cudaMemcpyHostToDevice, streams[1]));
        parallel_sum <<<(n*n/1024),1024,0,streams[1]>>> (d_rho1, d_t_rho1, *theta_max1, d_sum1);
        err_chk(cudaMemcpyAsync(sum1_2, d_sum1, sizeof(float), cudaMemcpyDeviceToHost, streams[1]));
      }
      
      if (norm_type == 1) {
        mUpdate_l1 << < n*n / threads_per_block, threads_per_block, 0, streams[2] >> > (d_Phi, d_m, d_m_temp, n, dx, mu);
      } else if (norm_type == 2) {
        mUpdate_l2 << < n*n / threads_per_block, threads_per_block, 0,  streams[2] >> > (d_Phi, d_m, d_m_temp, n, dx, mu);
      } else {
        assert(false);
      }

      
      bool last_sync0 = false, last_sync1 = false;
      
      cudaStreamSynchronize(streams[0]);
      if(*sum0 < gamma) {
        *theta_min0 -= rho0_max;
        *theta_min0 -= 0.01f*abs(*theta_min0); //0.01 is to avoid numerical precision issues
        *sum_min0 = gamma;
      } else {
        if (*sum0_2<=gamma) { //theta_max0 was valid
          //theta_max0 = theta_max0;
          *sum_max0 = *sum0_2;
          #ifndef NOTDEBUGFLAG 
          cout << "smart interval success0" << endl;
          #endif
        }
        else {
          //theta_max0 was not valid
          *theta_max0 = -1.0f*std::numeric_limits<float>::infinity(); //-inf
          err_chk(cudaMemcpyAsync(d_max0, theta_max0, sizeof(float), cudaMemcpyHostToDevice, streams[0]));
          cuda_parallel_max <<<(n*n/1024),1024,0,streams[0]>>> (d_t_rho0, d_max0);
          err_chk(cudaMemcpyAsync(theta_max0, d_max0, sizeof(float), cudaMemcpyDeviceToHost, streams[0]));
          last_sync0 = true;
        }
      }
      
      cudaStreamSynchronize(streams[1]);
      if (*sum1 < gamma) {
        *theta_min1 -= rho1_max;
        *theta_min1 -= 0.01f*abs(*theta_min1); //0.01 is to avoid numerical precision issues
        *sum_min1 = gamma;
      } else {

        if (*sum1_2<=gamma) { //theta_max1 was valid
          //*theta_max1 = *theta_max1;
          *sum_max1 = *sum1_2;
          #ifndef NOTDEBUGFLAG 
          cout << "smart interval success1" << endl;
          #endif
        }
        else {
          //theta_max1 was not valid
          *theta_max1 = -1.0f*std::numeric_limits<float>::infinity(); //-inf
          err_chk(cudaMemcpyAsync(d_max1, theta_max1, sizeof(float), cudaMemcpyHostToDevice, streams[1]));
          cuda_parallel_max <<<(n*n/1024),1024,0,streams[1]>>> (d_t_rho1, d_max1);
          err_chk(cudaMemcpyAsync(theta_max1, d_max1, sizeof(float), cudaMemcpyDeviceToHost, streams[1]));
          last_sync1 = true;
        }
      }
      
      
      if (last_sync0) {
        cudaStreamSynchronize(streams[0]);
        *sum_max0 = 0.0f;   
      }
      if (last_sync1) {
        cudaStreamSynchronize(streams[1]);
        *sum_max1 = 0.0f;   
      }
     
      
    }

    *theta_mid0 = (*theta_max0+*theta_min0)/2;
    *theta_mid1 = (*theta_max1+*theta_min1)/2;




    #ifndef NOTDEBUGFLAG 
    int counter0 = 0, counter1 = 0;
    #endif
    
    bool theta0_done = false, theta1_done = false;
    
    do {
      *sum0 = 0.0f;
      *sum1 = 0.0f;
      
      if(!theta0_done) {
        #ifndef NOTDEBUGFLAG 
        counter0++;
        #endif

        err_chk(cudaMemcpyAsync(d_sum0, sum0, sizeof(float), cudaMemcpyHostToDevice, streams[0]));
        parallel_sum <<<(n*n/1024),1024,0,streams[0]>>> (d_rho0, d_t_rho0, *theta_mid0, d_sum0);
        err_chk(cudaMemcpyAsync(sum0, d_sum0, sizeof(float), cudaMemcpyDeviceToHost, streams[0]));
      }
      if(!theta1_done) {
        #ifndef NOTDEBUGFLAG 
        counter1++;
        #endif
        err_chk(cudaMemcpyAsync(d_sum1, sum1, sizeof(float), cudaMemcpyHostToDevice, streams[1]));
        parallel_sum <<<(n*n/1024),1024,0,streams[1]>>> (d_rho1, d_t_rho1, *theta_mid1, d_sum1);
        err_chk(cudaMemcpyAsync(sum1, d_sum1, sizeof(float), cudaMemcpyDeviceToHost, streams[1]));
      }
      

      if(!theta0_done) {
        cudaStreamSynchronize(streams[0]);
        if (*sum0>gamma) {
          *theta_min0 = *theta_mid0;
          *sum_min0 = *sum0;
        }
        else {
          *theta_max0 = *theta_mid0;
          *sum_max0 = *sum0;
        }
        *theta_mid0 = (*theta_max0+*theta_min0)/2.f;
        theta0_done = !((*theta_max0-*theta_min0)>epsilon* abs(*theta_mid0) && (*sum_min0-*sum_max0)>0.001f );
      }
      if(!theta1_done) {
        cudaStreamSynchronize(streams[1]);
        if (*sum1>gamma) {
          *theta_min1 = *theta_mid1;
          *sum_min1 = *sum1;
        }
        else {
          *theta_max1 = *theta_mid1;
          *sum_max1 = *sum1;
        }
        *theta_mid1 = (*theta_max1+*theta_min1)/2.f;
        theta1_done = !((*theta_max1-*theta_min1)>epsilon* abs(*theta_mid1) && (*sum_min1-*sum_max1)>0.001f );
      }
    } while(!theta0_done || !theta1_done);
    
    if (use_smart_interval) {
      theta_deque0.push_back(*theta_mid0);
      theta_deque1.push_back(*theta_mid1);
      if (k>=DEQUE_SIZE) {
        theta_deque0.pop_front();
        theta_deque1.pop_front();
      }
    }
    
    
    #ifndef NOTDEBUGFLAG 
    cout << counter0 << " bisection iterations for theta0" << endl;
    cout << counter1 << " bisection iterations for theta1" << endl;
    #endif
     
    
    

    rho_commit <<< n*n / threads_per_block, threads_per_block, 0, streams[0] >>> (d_rho0, d_t_rho0, d_t_rho0_temp, *theta_mid0);
    rho_commit <<< n*n / threads_per_block, threads_per_block, 0, streams[1] >>> (d_rho1, d_t_rho1, d_t_rho1_temp, *theta_mid1);

    //End rho updates
    //----------------------------------------
    

    //Phi = Phi + tau*(div m_temp-t_rho0_temp+t_rho1_temp)
    PhiUpdate <<< n*n / threads_per_block, threads_per_block >>> (d_Phi, d_m_temp, d_t_rho0_temp, d_t_rho1_temp, n, dx, tau);
    cudaDeviceSynchronize();
    //err_chk(cudaGetLastError());
    
  }
  cudaDeviceSynchronize();
  float runtime = float(clock() - begin) / CLOCKS_PER_SEC;
  cout << "Total runtime is " << runtime << "s." << endl;
  cout << "This is " << runtime / float(max_iter) * 1000 << "ms per iteration" << endl;


  cout << "copying m and Phi from GPU to host" << endl;
  err_chk(cudaMemcpy(m, d_m, sizeof(float)*n*n*d, cudaMemcpyDeviceToHost));
  err_chk(cudaMemcpy(Phi, d_Phi, sizeof(float)*n*n, cudaMemcpyDeviceToHost));
  err_chk(cudaMemcpy(t_rho0, d_t_rho0, sizeof(float)*n*n, cudaMemcpyDeviceToHost));
  err_chk(cudaMemcpy(t_rho1, d_t_rho1, sizeof(float)*n*n, cudaMemcpyDeviceToHost));


  
  
  
  
  
  
  
  if (norm_type == 1) {
    plhs[0] = mxCreateDoubleScalar(wass_l1(m, n, d));
  } else if (norm_type == 2) {
    plhs[0] = mxCreateDoubleScalar(wass_l2(m, n, d));
  } else {
    assert(false);
  }
  if (nlhs==5) {
    size_t m_dim[3] = {n,n,d};
    size_t phi_dim[2] = {n,n};
    size_t t_rho0_dim[2] = {n,n};
    size_t t_rho1_dim[2] = {n,n};
    plhs[1] = mxCreateNumericArray(3, m_dim, mxDOUBLE_CLASS, mxREAL);
    plhs[2] = mxCreateNumericArray(2, phi_dim, mxDOUBLE_CLASS, mxREAL);
    plhs[3] = mxCreateNumericArray(2, t_rho0_dim, mxDOUBLE_CLASS, mxREAL);
    plhs[4] = mxCreateNumericArray(2, t_rho1_dim, mxDOUBLE_CLASS, mxREAL);
    
    
    double* m_out = mxGetPr(plhs[1]);
    double* phi_out = mxGetPr(plhs[2]);
    double* t_rho0_out = mxGetPr(plhs[3]);
    double* t_rho1_out = mxGetPr(plhs[4]);
    
    for (int ii=0; ii<n*n*d; ii++)
      m_out[ii] = double(m[ii]);
    
    for (int ii=0; ii<n*n; ii++)
      phi_out[ii] = double(Phi[ii]);
      
    for (int ii=0; ii<n*n; ii++)
      t_rho0_out[ii] = double(t_rho0[ii]);
      
    for (int ii=0; ii<n*n; ii++)
      t_rho1_out[ii] = double(t_rho1[ii]);
    
  }
  
  
  
  
  cudaStreamDestroy(streams[0]);
  cudaStreamDestroy(streams[1]);
  cudaStreamDestroy(streams[2]);
  

  cout << "freeing CUDA resources" << endl;
  cudaFree(d_rho0);
  cudaFree(d_rho1);
  cudaFree(d_m);
  cudaFree(d_m_temp);
  cudaFree(d_Phi);
  cudaFree(d_t_rho0);
  cudaFree(d_t_rho0_temp);
  cudaFree(d_t_rho1);
  cudaFree(d_t_rho1_temp);
  cudaFree(d_min0);
  cudaFree(d_max0);
  cudaFree(d_sum0);
  cudaFree(d_min1);
  cudaFree(d_max1);
  cudaFree(d_sum1);
  
  
  cudaFreeHost(theta_min0);
  cudaFreeHost(theta_max0);
  cudaFreeHost(theta_mid0);
  cudaFreeHost(sum_min0);
  cudaFreeHost(sum_max0);  
  cudaFreeHost(sum0);  
  cudaFreeHost(sum0_2);  
  cudaFreeHost(theta_min1);
  cudaFreeHost(theta_max1);
  cudaFreeHost(theta_mid1);
  cudaFreeHost(sum_min1);
  cudaFreeHost(sum_max1);  
  cudaFreeHost(sum1);  
  cudaFreeHost(sum1_2);  


  cout << "freeing host resources" << endl;
  delete[] rho0;
  delete[] rho1;
  delete[] m;
  delete[] Phi;
  delete[] m_prev;
  delete[] Phi_prev;
  delete[] t_rho0;
  delete[] t_rho1;
  delete[] t_rho0_temp;
  delete[] t_rho1_temp;
  delete[] t_rho0_prev;
  delete[] t_rho1_prev;
}







