#include <cassert>
#include <string>
#include <ctime>
#include <cmath>
#include <deque>
#include <limits>
#include "somt.h"
#include "EMD_header.h"

using std::cout;
using std::endl;
using wtd::deque;



#define WARP_SIZE 32



class mystream : public std::streambuf
{
protected:
virtual std::streamsize xsputn(const char *s, std::streamsize n) { mexPrintf("%.*s", n, s); return n; }
virtual int overflow(int c=EOF) { if (c != EOF) { mexPrintf("%.1s", &c); } return 1; }
};
class scoped_redirect_cout
{
public:
	scoped_redirect_cout() { old_buf = std::cout.rdbuf(); std::cout.rdbuf(&mout); }
	~scoped_redirect_cout() { std::cout.rdbuf(old_buf); }
private:
	mystream mout;
	std::streambuf *old_buf;
};
static scoped_redirect_cout mycout_redirect;





//CUDA kernels
__global__ void lambdaStep(float* d_Phi, float* d_t_lambda0, float* d_t_lambda0_temp, float* d_t_lambda1, float* d_t_lambda1_temp, int nu, int n);
inline void lambdaUpdate(float* d_Phi, float* d_lambda0, float* d_lambda1, float* d_t_lambda0, float* d_t_lambda0_temp, float* d_t_lambda1, float* d_t_lambda1_temp, float nu, float gamma, float lambda0_max, float lambda1_max, int n, int threads_per_block, float* d_min, float* d_max, float* d_sum);
__global__ void cuda_parallel_minmax(float *lambda0, int num_elements, float *max, float *min);
__global__ void parallel_sum(float *lambda, float* t_lambda, int num_elements, float theta_mid, float *sum);
__global__ void lambda_commit(float *lambda, float* t_lambda, float* t_lambda_temp, int num_elements, float theta_mid);




void mexFunction(int nlhs, mxArray *plhs[], int nrhs, mxArray const *prhs[]) {
  const int d = 2;
  float mu;
  float nu;
  float tau;
  int max_iter;
  bool use_smart_interval;
  int norm_type;
  
  mu = float(*mxGetPr(prhs[2]));
  mu = float(*mxGetPr(prhs[3]));
  tau = float(*mxGetPr(prhs[4]));
  max_iter = int(*mxGetPr(prhs[5]));
  use_smart_interval mxIsLogicalScalarTrue( prhs[6] );
  norm_type = int(*mxGetPr(prhs[7]));


  
  if ( nrhs!=6 ){
    mexErrMsgIdAndTxt("error:error",
                      "wrong number of inputs.");
  }
  
  if ( !(nlhs==1 || nlhs==5) ) {
    mexErrMsgIdAndTxt("error:error",
                      "output should be dist or [dist U Phi]");
  }

  const int n = mxGetDimensions(prhs[0])[0];
  const float dx = 1. / float(n - 1);
  
  const double* lambda0_float = mxGetPr(prhs[0]);
  const double* lambda1_float = mxGetPr(prhs[1]);
  
  float* lambda0 = new float[n*n];
  float* lambda1 = new float[n*n];
  
  //convert input to float
  for (int i = 0; i < n; i++) {
    for (int j = 0; j < n; j++) {
      lambda0[i + n*j] = float(lambda0_float[i + n*j]);
      lambda1[i + n*j] = float(lambda1_float[i + n*j]);
    }
  }

  
  if ( !( n==1 || n==2 || n==4 || n==8 || n==16 || n==32 || n==64 || n==128 || n==256) ) {
    mexErrMsgIdAndTxt("error:error",
                      "n has to be a power of 2");
  }
  
  
  
  int threads_per_block;
  if (n==1) {
    threads_per_block = 1;
  } else if (n==2) {
    threads_per_block = 4;
  } else if (n==4) {
    threads_per_block = 16;
  } else if (n==8) {
    threads_per_block = 64;
  } else if (n==16) {
    threads_per_block = 128;
  } else if (n==32) {
    threads_per_block = 128;
  } else if (n==64) {
    threads_per_block = 256;
  } else if (n==128) {
    threads_per_block = 256;
  } else if (n==256) {
    threads_per_block = 256;
  } else {
    mexErrMsgIdAndTxt("error:error",
                      "unknown error");
  }
  
  



  float* t_lambda0 = new float[n*n];
  float* t_lambda1 = new float[n*n];
  float* t_lambda0_temp = new float[n*n];
  float* t_lambda1_temp = new float[n*n];  

  for (int ii=0; ii<n*n; ii++) {
    t_lambda0[ii] = lambda0[ii];
    t_lambda1[ii] = lambda1[ii];
  }


  //initialize m and Phi to be 0
  float* m = new float[n*n*d];
  memset(m, 0, sizeof(float)*n*n*d);
  float* Phi = new float[n*n];
  memset(Phi, 0, sizeof(float)*n*n);
  float* m_prev = new float[n*n*d];
  float* Phi_prev = new float[n*n];
  float* t_lambda0_prev = new float[n*n];
  float* t_lambda1_prev = new float[n*n];


  

  //---------------------------------------------------------------------
  //start error checking
  //thread_per_block must divide number of points
  assert((n*n) % threads_per_block == 0);
  if (n*n / threads_per_block > 65535) {
    cout << "maximum block number exceeded" << endl;
    cout << "threads_per_block should be bigger" << endl;
    assert(false);
  }
  assert((n*n)%1024==0); //required for parallel reduction
  //end error checking
  //---------------------------------------------------------------------

  //create CUDA pointers
  float* d_lambda0;
  float* d_lambda1;
  float* d_m;
  float* d_m_temp;
  float* d_Phi;
  float* d_t_lambda0;
  float* d_t_lambda0_temp;
  float* d_t_lambda1;
  float* d_t_lambda1_temp;
  float* d_min;
  float* d_max;
  float* d_sum;

  cout << "Allocating GPU memory" << endl;
  err_chk(cudaMalloc((void**)&d_lambda0, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_lambda1, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_m, sizeof(float)*n*n*d));
  err_chk(cudaMalloc((void**)&d_m_temp, sizeof(float)*n*n*d));
  err_chk(cudaMalloc((void**)&d_Phi, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_lambda0, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_lambda0_temp, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_lambda1, sizeof(float)*n*n));
  err_chk(cudaMalloc((void**)&d_t_lambda1_temp, sizeof(float)*n*n));

  err_chk(cudaMalloc((void**)&d_min, sizeof(float)));
  err_chk(cudaMalloc((void**)&d_max, sizeof(float)));
  err_chk(cudaMalloc((void**)&d_sum, sizeof(float)));


  cout << "copying memory from host to GPU" << endl;
  err_chk(cudaMemcpy(d_m, m, sizeof(float)*n*n*d, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_Phi, Phi, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_lambda0, lambda0, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_lambda1, lambda1, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_t_lambda0, t_lambda0, sizeof(float)*n*n, cudaMemcpyHostToDevice));
  err_chk(cudaMemcpy(d_t_lambda1, t_lambda1, sizeof(float)*n*n, cudaMemcpyHostToDevice));



  deque<float> theta_deque1;
  deque<float> theta_deque2;


  cout << "starting iteration" << endl;
  clock_t begin = clock();
  for (int k = 0; k < max_iter; k++) {
  
    if (norm_type == 1) {
      mUpdate_l1 << < n*n / threads_per_block, threads_per_block >> > (d_Phi, d_m, d_m_temp, n, dx, mu);
    } else if (norm_type == 2) {
      mUpdate_l2 << < n*n / threads_per_block, threads_per_block >> > (d_Phi, d_m, d_m_temp, n, dx, mu);
    } else {
      assert(false);
    }
    
    
    lambdaUpdate(d_Phi, d_lambda0, d_lambda1, d_t_lambda0, d_t_lambda0_temp, d_t_lambda1, d_t_lambda1_temp, nu, gamma, lambda0_max, lambda1_max, n, threads_per_block, d_min, d_max, d_sum);

    //Phi = Phi + tau*(div m_temp-t_lambda0_temp+t_lambda1_temp)
    PhiUpdate << < n*n / threads_per_block, threads_per_block >> >(d_Phi, d_m_temp, d_t_lambda0_temp, d_t_lambda1_temp, n, dx, tau);

    //err_chk(cudaGetLastError());
    
  }
  cudaDeviceSynchronize();
  float runtime = float(clock() - begin) / CLOCKS_PER_SEC;
  cout << "Total runtime is " << runtime << "s." << endl;
  cout << "This is " << runtime / float(max_iter) * 1000 << "ms per iteration" << endl;


  cout << "copying m and Phi from GPU to host" << endl;
  err_chk(cudaMemcpy(m, d_m, sizeof(float)*n*n*d, cudaMemcpyDeviceToHost));
  err_chk(cudaMemcpy(Phi, d_Phi, sizeof(float)*n*n, cudaMemcpyDeviceToHost));
  err_chk(cudaMemcpy(t_lambda0, d_t_lambda0, sizeof(float)*n*n, cudaMemcpyDeviceToHost));
  err_chk(cudaMemcpy(t_lambda1, d_t_lambda1, sizeof(float)*n*n, cudaMemcpyDeviceToHost));



  cout << "freeing CUDA resources" << endl;
  cudaFree(d_lambda0);
  cudaFree(d_lambda1);
  cudaFree(d_m);
  cudaFree(d_m_temp);
  cudaFree(d_Phi);
  cudaFree(d_t_lambda0);
  cudaFree(d_t_lambda0_temp);
  cudaFree(d_t_lambda1);
  cudaFree(d_t_lambda1_temp);
  cudaFree(d_min);
  cudaFree(d_max);
  cudaFree(d_sum);


  cout << "freeing host resources" << endl;
  delete[] lambda0;
  delete[] lambda1;
  delete[] m;
  delete[] Phi;
  delete[] m_prev;
  delete[] Phi_prev;
  delete[] t_lambda0;
  delete[] t_lambda1;
  delete[] t_lambda0_temp;
  delete[] t_lambda1_temp;
  delete[] t_lambda0_prev;
  delete[] t_lambda1_prev;
}




inline void lambdaUpdate(float* d_Phi, float* d_lambda0, float* d_lambda1, float* d_t_lambda0, float* d_t_lambda0_temp, float* d_t_lambda1, float* d_t_lambda1_temp, float nu, float gamma, float lambda0_max, float lambda1_max, int n, int threads_per_block, float* d_min, float* d_max, float* d_sum) {

  const float epsilon = 1e-6;
  
  lambdaStep << < n*n / threads_per_block, threads_per_block >> > (d_Phi, d_t_lambda0, d_t_lambda0_temp, d_t_lambda1, d_t_lambda1_temp, nu, n);
  
  
  float theta_min, theta_max, theta_mid;

  
  //compute theta range
  if (!use_smart_interval || k==0) {
    theta_min = std::numeric_limits<float>::infinity(); //inf
    theta_max = -1*std::numeric_limits<float>::infinity(); //-inf
    err_chk(cudaMemcpy(d_min, &theta_min, sizeof(float), cudaMemcpyHostToDevice));
    err_chk(cudaMemcpy(d_max, &theta_max, sizeof(float), cudaMemcpyHostToDevice));
    cuda_parallel_minmax <<<(n*n/1024),1024>>> (d_t_lambda0, n*n, d_max, d_min);
    err_chk(cudaMemcpy(&theta_max, d_max, sizeof(float), cudaMemcpyDeviceToHost));
    err_chk(cudaMemcpy(&theta_min, d_min, sizeof(float), cudaMemcpyDeviceToHost));
    theta_min -= lambda0_max;
    theta_mid = (theta_max+theta_min)/2;
    sum_min = gamma;
    sum_max = 0.0f;
    
  } else {
  
    float sum = 0.0f;
    err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
    parallel_sum <<<(n*n/1024),1024>>> (d_rho0, d_t_rho0, n*n, theta_mid_prev0 - C*epsilon*abs(theta_mid_prev0), d_sum);
    err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));
    if (sum > 0.99*gamma && sum < 1.01*gamma) {
      theta_min = theta_max = theta_mid_prev1 - C*epsilon*abs(theta_mid_prev1);
    }
    else if (sum < gamma){
      theta_max = theta_mid_prev0 - C*epsilon*abs(theta_mid_prev0);
      sum_max = sum;
      float temp_min = 1.0/0.0; //inf
      err_chk(cudaMemcpy(d_min, &temp_min, sizeof(float), cudaMemcpyHostToDevice));
      cuda_parallel_min <<<(n*n/1024),1024>>> (d_t_rho0, n*n, d_min);
      err_chk(cudaMemcpy(&temp_min, d_min, sizeof(float), cudaMemcpyDeviceToHost));
      theta_min = temp_min - rho0_max;
      theta_min -= 0.1*abs(theta_min); //0.1 is to avoid numerical precision issues
      sum_min = gamma;
    }
    else {
      theta_min = theta_mid_prev0 - C*epsilon*abs(theta_mid_prev0);
      sum_min = sum;

      sum = 0.0f;
      err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
      parallel_sum <<<(n*n/1024),1024>>> (d_rho0, d_t_rho0, n*n, theta_mid_prev0 + C*epsilon*abs(theta_mid_prev0), d_sum);
      err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));

      if (sum<=gamma) {
        theta_max = theta_mid_prev0 + C*epsilon*abs(theta_mid_prev0);
      }
      else {
        float temp_max = -1.0/0.0; //-inf
        err_chk(cudaMemcpy(d_max, &temp_max, sizeof(float), cudaMemcpyHostToDevice));
        cuda_parallel_max <<<(n*n/1024),1024>>> (d_t_rho0, n*n, d_max);
        err_chk(cudaMemcpy(&temp_max, d_max, sizeof(float), cudaMemcpyDeviceToHost));
        theta_max = temp_max;
        sum_max = 0.0f;
      }
    }

  }

  
  

  do {
    float sum = 0.0f;
    err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
    parallel_sum <<<(n*n/1024),1024>>> (d_rho0, d_t_rho0, n*n, theta_mid, d_sum);
    err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));

    if (sum>gamma) {
      theta_min = theta_mid;
      sum_min = sum;
    }
    else {
      theta_max = theta_mid;
      sum_max = sum;
    }
    theta_mid = (theta_max+theta_min)/2;
  }  while((theta_max-theta_min)>epsilon* abs(theta_mid) && (sum_min-sum_max)>0.001 );

  if (use_smart_interval) {
    theta_deque1.push_back(theta_mid);
    if (k>=10)
      theta_deque1.pop_front(theta_mid);
  }

  
  
  lambda_commit << < n*n / threads_per_block, threads_per_block >> > (d_lambda0, d_t_lambda0, d_t_lambda0_temp, n*n, theta_mid);

  
  if (!use_smart_interval || k==0) {
    theta_min = 1.0f/0.0f; //inf
    theta_max = -1.0f/0.0f; //-inf
    err_chk(cudaMemcpy(d_min, &theta_min, sizeof(float), cudaMemcpyHostToDevice));
    err_chk(cudaMemcpy(d_max, &theta_max, sizeof(float), cudaMemcpyHostToDevice));
    cuda_parallel_minmax <<<(n*n/1024),1024>>> (d_t_lambda1, n*n, d_max, d_min);
    err_chk(cudaMemcpy(&theta_max, d_max, sizeof(float), cudaMemcpyDeviceToHost));
    err_chk(cudaMemcpy(&theta_min, d_min, sizeof(float), cudaMemcpyDeviceToHost));
    theta_min -= lambda1_max;
    theta_mid = (theta_max+theta_min)/2;
    sum_min = gamma;
    sum_max = 0.0f;
  } else {
  
    float sum = 0.0f;
    err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
    parallel_sum <<<(n*n/1024),1024>>> (d_rho0, d_t_rho0, n*n, theta_mid_prev0 - C*epsilon*abs(theta_mid_prev0), d_sum);
    err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));
    if (sum > 0.99*gamma && sum < 1.01*gamma) {
      theta_min = theta_max = theta_mid_prev1 - C*epsilon*abs(theta_mid_prev1);
    }
    else if (sum < gamma){
      theta_max = theta_mid_prev0 - C*epsilon*abs(theta_mid_prev0);
      sum_max = sum;
      float temp_min = 1.0/0.0; //inf
      err_chk(cudaMemcpy(d_min, &temp_min, sizeof(float), cudaMemcpyHostToDevice));
      cuda_parallel_min <<<(n*n/1024),1024>>> (d_t_rho0, n*n, d_min);
      err_chk(cudaMemcpy(&temp_min, d_min, sizeof(float), cudaMemcpyDeviceToHost));
      theta_min = temp_min - rho0_max;
      theta_min -= 0.1*abs(theta_min); //0.1 is to avoid numerical precision issues
      sum_min = gamma;
    }
    else {
      theta_min = theta_mid_prev0 - C*epsilon*abs(theta_mid_prev0);
      sum_min = sum;

      sum = 0.0f;
      err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
      parallel_sum <<<(n*n/1024),1024>>> (d_rho0, d_t_rho0, n*n, theta_mid_prev0 + C*epsilon*abs(theta_mid_prev0), d_sum);
      err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));

      if (sum<=gamma) {
        theta_max = theta_mid_prev0 + C*epsilon*abs(theta_mid_prev0);
      }
      else {
        float temp_max = -1.0/0.0; //-inf
        err_chk(cudaMemcpy(d_max, &temp_max, sizeof(float), cudaMemcpyHostToDevice));
        cuda_parallel_max <<<(n*n/1024),1024>>> (d_t_rho0, n*n, d_max);
        err_chk(cudaMemcpy(&temp_max, d_max, sizeof(float), cudaMemcpyDeviceToHost));
        theta_max = temp_max;
        sum_max = 0.0f;
      }
    }

  }

  
  
  
  
  theta_mid = (theta_max+theta_min)/2;
  do {
    //cout << "counter 1" << endl;
    float sum = 0.0f;
    err_chk(cudaMemcpy(d_sum, &sum, sizeof(float), cudaMemcpyHostToDevice));
    parallel_sum <<<(n*n/1024),1024>>> (d_rho1, d_t_rho1, n*n, theta_mid, d_sum);
    err_chk(cudaMemcpy(&sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost));

    if (sum>gamma) {
      theta_min = theta_mid;
      sum_min = sum;
    }
    else {
      theta_max = theta_mid;
      sum_max = sum;
    }
    theta_mid = (theta_max+theta_min)/2;
   }  while((theta_max-theta_min)>epsilon* abs(theta_mid) && (sum_min-sum_max)>0.001 );


   
  if (use_smart_interval) {
    theta_deque2.push_back(theta_mid);
    if (k>=10)
      theta_deque2.pop_front(theta_mid);
  }


   lambda_commit << < n*n / threads_per_block, threads_per_block >> > (d_lambda1, d_t_lambda1, d_t_lambda1_temp, n*n, theta_mid);



}


lambdaupdate2 }

}





__device__ void fatomicMax(float* max, float val) {
  //there is no atomic add, so we use atomicCAS
  unsigned int* address_as_ui = (unsigned int*) max;
  unsigned int old = *address_as_ui;
  unsigned int assumed;

  do {
    assumed = old;
    old = atomicCAS(address_as_ui, assumed,
     __float_as_int( fmaxf(val, __int_as_float(assumed)) ));

  } while(assumed != old);
}

__device__ void fatomicMin(float* min, float val) {
  //there is no atomic add, so we use atomicCAS
  unsigned int* address_as_ui = (unsigned int*) min;
  unsigned int old = *address_as_ui;
  unsigned int assumed;

  do {
    assumed = old;
    old = atomicCAS(address_as_ui, assumed,
     __float_as_int( fminf(val, __int_as_float(assumed))) );

  } while(assumed != old);
}


//blocksize must be 1024
// num_elemts must be a multiple of 1024
__global__ void cuda_parallel_minmax(float *lambda, int num_elements, float *max, float *min) {


  //Holds intermediates in shared memory reduction
  __syncthreads();

  __shared__ float buffer[1024/WARP_SIZE];

  int globalIdx = blockIdx.x * blockDim.x + threadIdx.x;
  int lane = threadIdx.x % WARP_SIZE;


  float temp;
  // All threads in a block of 1024 take an element
  temp = lambda[globalIdx];
  
  // All warps in this block (32) compute the sum of all
  // threads in their warp
  for(int delta = WARP_SIZE/2; delta > 0; delta /= 2)
    temp = fmaxf(temp, __shfl_xor(temp, delta));
  
  // Write all 32 of these partial sums to shared memory
  if(lane == 0)
    buffer[threadIdx.x / WARP_SIZE] = temp;

  __syncthreads();
  

  // Add the remaining 32 partial sums using a single warp
  if(threadIdx.x < WARP_SIZE) {
    temp = buffer[threadIdx.x];
    for(int delta = WARP_SIZE / 2; delta > 0; delta /= 2)
      temp = fmaxf(temp, __shfl_xor(temp, delta));
  }

  // Add this block's sum to the total sum
  if(threadIdx.x == 0)
    fatomicMax(max, temp);
  

  __syncthreads();
  // All threads in a block of 1024 take an element
  temp = lambda[globalIdx];
    
  // All warps in this block (32) compute the sum of all
  // threads in their warp
  for(int delta = WARP_SIZE/2; delta > 0; delta /= 2)
    temp = fminf(temp, __shfl_xor(temp, delta));

    
  // Write all 32 of these partial sums to shared memory
  if(lane == 0)
    buffer[threadIdx.x / WARP_SIZE] = temp;

  __syncthreads();


  // Add the remaining 32 partial sums using a single warp
  if(threadIdx.x < WARP_SIZE) {
    temp = buffer[threadIdx.x];
    for(int delta = WARP_SIZE / 2; delta > 0; delta /= 2)
      temp = fminf(temp, __shfl_xor(temp, delta));
  }

  // Add this block's sum to the total sum
  if(threadIdx.x == 0)
    fatomicMin(min, temp);

}




__global__ void parallel_sum(float *lambda, float* t_lambda, int num_elements, float theta_mid, float *sum) {
  //Holds intermediates in shared memory reduction
  __syncthreads();

  __shared__ float buffer[1024/WARP_SIZE];
  int globalIdx = blockIdx.x * blockDim.x + threadIdx.x;
  int lane = threadIdx.x % WARP_SIZE;


  float temp;
  // All threads in a block of 1024 take an element
  temp = fminf(lambda[globalIdx], fmaxf(t_lambda[globalIdx]-theta_mid,0.0f));

  // All warps in this block (32) compute the sum of all
  // threads in their warp
  for(int delta = WARP_SIZE/2; delta > 0; delta /= 2)
    temp += __shfl_xor(temp, delta);


  // Write all 32 of these partial sums to shared memory
  if(lane == 0)
    buffer[threadIdx.x / WARP_SIZE] = temp;
    
  __syncthreads();


  // Add the remaining 32 partial sums using a single warp
  if(threadIdx.x < WARP_SIZE) {
    temp = buffer[threadIdx.x];
    for(int delta = WARP_SIZE / 2; delta > 0; delta /= 2)
        temp += __shfl_xor(temp, delta);
  }

  // Add this block's sum to the total sum
  if(threadIdx.x == 0)
    atomicAdd(sum, temp);

}






__global__ void lambda_commit(float *lambda, float* t_lambda, float* t_lambda_temp, int num_elements, float theta_mid) {
  int ii = blockDim.x*blockIdx.x + threadIdx.x;
  t_lambda[ii] = fminf(lambda[ii], fmaxf(t_lambda[ii]-theta_mid,0.0f));
  t_lambda_temp[ii] = 2*t_lambda[ii] - t_lambda_temp[ii];  
}



__global__ void lambdaStep(float* d_Phi, float* d_t_lambda0, float* d_t_lambda0_temp, float* d_t_lambda1, float* d_t_lambda1_temp, int nu, int n) {
  //int N = n*n;
  int ii = blockDim.x*blockIdx.x + threadIdx.x;
  //int i = ii % n;
  //int j = ii / n;

  d_t_lambda0_temp[ii] = d_t_lambda0[ii];
  d_t_lambda0[ii] += nu*d_Phi[ii];

  d_t_lambda1_temp[ii] = d_t_lambda1[ii];
  d_t_lambda1[ii] -= nu*d_Phi[ii];
}





